unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, SyncObjs, Forms, Controls, Graphics, Dialogs, StdCtrls;

type

  TThreadSimple = class(TThread)
    private

    protected
      procedure Execute; override;

    public
      constructor Create(CreateSuspended: boolean);
      destructor Destroy; override;
    end;

  TThreadEvent = class(TThread)
    private
     const
     FMaxLockCount = 10;
    private
      FShuttingDown: Boolean;  // Avoid to wait when killing the Thread. Terminated only is slow.
      FEvent: TEvent;
      FLockCount: Integer;

    protected
      procedure Execute; override;

    public
      constructor Create(CreateSuspended: boolean);
      destructor Destroy; override;
      procedure Kill;
    end;



  { TForm1 }

  TForm1 = class(TForm)
    BtnStartThread1: TButton;
    BtnStartThread2: TButton;
    BtnStartThread3: TButton;
    BtnStartThread4: TButton;
    procedure BtnStartThread1Click(Sender: TObject);
    procedure BtnStartThread2Click(Sender: TObject);
    procedure BtnStartThread3Click(Sender: TObject);
    procedure BtnStartThread4Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    MyThread1: TThreadSimple;
    MyThread2: TThreadEvent;
    MyThread3: TThreadEvent;
    MyThread4: TThreadEvent;

    procedure StartThread(aThread: TThread; aBtn: TButton; aName: String);
    procedure KillThread(aThread: TThread);
  public
  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}

{ TMyThread1 }
constructor TThreadSimple.Create(CreateSuspended: boolean);
begin
  Inherited Create(CreateSuspended);
end;

destructor TThreadSimple.Destroy;
begin
  inherited Destroy;
end;

procedure TThreadSimple.Execute;
begin
  if Terminated
    then Exit;

  Try
     // DoSomething
  Except
    On Exception do;
  end;
end;


{ TThreadEvent }

constructor TThreadEvent.Create(CreateSuspended: boolean);
begin
  FEvent:= TEvent.Create(nil, False, False, '');
  Inherited Create(CreateSuspended);
  FShuttingDown := False;
end;

destructor TThreadEvent.Destroy;
begin
  FEvent.Free;
  inherited Destroy;
end;

procedure TThreadEvent.Kill;
begin
  { send event to stop the thread execute proc }
  FEvent.SetEvent;
end;

procedure TThreadEvent.Execute;
var
  aSleepDelay: Integer = 0;
begin
  FLockCount:= FMaxLockCount;
  While not Terminated do
  Try
    Try
    Case FEvent.WaitFor(aSleepDelay) of
      wrSignaled : begin
                   FShuttingDown:= True;  // Avoid to wait when killing the Thread. Terminated only is slow.
                   Terminate; { Event to terminate the thread was set }
                   end;
      wrTimeout  : begin
                   if FShuttingDown
                     then Kill;
                   // DoSometing
                   end;
     end;
   Except
     On Exception do; // DestroyMutex(agmAutoSyncData);
   end;
  finally
   Sleep(aSleepDelay);
  end;
end;


{ TForm1 }

procedure TForm1.FormCreate(Sender: TObject);
begin
  MyThread1 := TThreadSimple.Create(True);
  MyThread2 := TThreadEvent.Create(True);
  MyThread3 := TThreadEvent.Create(True);
  MyThread4 := TThreadEvent.Create(True);
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  KillThread(MyThread1);
  KillThread(MyThread2);
  KillThread(MyThread3);
  KillThread(MyThread4);
end;

procedure TForm1.StartThread(aThread: TThread; aBtn: TButton; aName: String);
begin
  if Assigned(aThread) and aThread.Suspended then
    begin
      aThread.Start;
      if aThread is TThreadSimple
        then aName := aName + ' (simple)'
        else aName := aName + ' (TEvent)';
      aBtn.Caption := aName + ' is running';
    end;
end;

procedure TForm1.KillThread(aThread: TThread);
begin
  if Assigned(aThread) then
  Try
    if aThread is TThreadEvent
      then TThreadEvent(aThread).Kill;
    if Not aThread.Suspended
      then aThread.WaitFor;
  finally
    aThread.Free;
  end;
end;

procedure TForm1.BtnStartThread1Click(Sender: TObject);
begin
  StartThread(MyThread1, TButton(Sender), 'MyThread1');
end;

procedure TForm1.BtnStartThread2Click(Sender: TObject);
begin
  StartThread(MyThread2, TButton(Sender), 'MyThread2');
end;

procedure TForm1.BtnStartThread3Click(Sender: TObject);
begin
  StartThread(MyThread3, TButton(Sender), 'MyThread3');
end;

procedure TForm1.BtnStartThread4Click(Sender: TObject);
begin
  StartThread(MyThread4, TButton(Sender), 'MyThread4');
end;


end.

